@@ -13,6 +13,7 @@ extern "C" {
 #endif
 
 #define SEXP_MODULE_PATH_VAR "CHIBI_MODULE_PATH"
+#define SEXP_NO_SYSTEM_PATH_VAR "CHIBI_IGNORE_SYSTEM_PATH"
 
 #include "chibi/features.h"
 #include "chibi/install.h"
@@ -81,6 +82,12 @@ typedef long long off_t;
 #define exit(x)           exits(TOSTRING(x))
 #define fabsl          fabs
 #define M_LN10         2.30258509299404568402  /* log_e 10 */
+#define FLT_RADIX 2
+#define isfinite(x) !(isNaN(x) || isInf(x,0))
+typedef u32int uint32_t;
+typedef s32int int32_t;
+typedef u64int uint64_t;
+typedef s64int int64_t;
 #else
 #include <stddef.h>
 #include <stdlib.h>
@@ -162,13 +169,13 @@ enum sexp_types {
   SEXP_VECTOR,
   SEXP_FLONUM,
   SEXP_BIGNUM,
-#if SEXP_USE_RATIOS
+#if SEXP_USE_STABLE_ABI || SEXP_USE_RATIOS
   SEXP_RATIO,
 #endif
-#if SEXP_USE_COMPLEX
+#if SEXP_USE_STABLE_ABI || SEXP_USE_COMPLEX
   SEXP_COMPLEX,
 #endif
-#if SEXP_USE_DISJOINT_STRING_CURSORS
+#if SEXP_USE_STABLE_ABI || SEXP_USE_DISJOINT_STRING_CURSORS
   SEXP_STRING_CURSOR,
 #endif
   SEXP_IPORT,
@@ -181,7 +188,7 @@ enum sexp_types {
   SEXP_ENV,
   SEXP_BYTECODE,
   SEXP_CORE,
-#if SEXP_USE_DL
+#if SEXP_USE_STABLE_ABI || SEXP_USE_DL
   SEXP_DL,
 #endif
   SEXP_OPCODE,
@@ -195,10 +202,11 @@ enum sexp_types {
   SEXP_STACK,
   SEXP_CONTEXT,
   SEXP_CPOINTER,
-#if SEXP_USE_AUTO_FORCE
+  SEXP_UNIFORM_VECTOR,
+#if SEXP_USE_STABLE_ABI || SEXP_USE_AUTO_FORCE
   SEXP_PROMISE,
 #endif
-#if SEXP_USE_WEAK_REFERENCES
+#if SEXP_USE_STABLE_ABI || SEXP_USE_WEAK_REFERENCES
   SEXP_EPHEMERON,
 #endif
   SEXP_NUM_CORE_TYPES
@@ -209,37 +217,51 @@ enum sexp_types {
 #endif
 
 #ifdef _WIN32
-#if defined(_MSC_VER) && SEXP_64_BIT
-/* On SEXP_64_BIT, 128bits arithmetic is mandatory */
-#error Unsupported configuration
-#endif
 #if SEXP_64_BIT
 typedef unsigned int sexp_tag_t;
 typedef unsigned long long sexp_uint_t;
 typedef long long sexp_sint_t;
+#define SEXP_PRIdFIXNUM "lld"
 #else
 typedef unsigned short sexp_tag_t;
 typedef unsigned int sexp_uint_t;
 typedef int sexp_sint_t;
+#define SEXP_PRIdFIXNUM "d"
 #endif
 #define sexp_heap_align(n) sexp_align(n, 5)
 #define sexp_heap_chunks(n) (sexp_heap_align(n)>>5)
 #elif SEXP_64_BIT
+#if PLAN9
+typedef uintptr sexp_tag_t;
+typedef uintptr sexp_uint_t;
+typedef intptr sexp_sint_t;
+#else
 typedef unsigned int sexp_tag_t;
 typedef unsigned long sexp_uint_t;
 typedef long sexp_sint_t;
+#endif
+#define SEXP_PRIdFIXNUM "ld"
 #define sexp_heap_align(n) sexp_align(n, 5)
 #define sexp_heap_chunks(n) (sexp_heap_align(n)>>5)
 #elif defined(__CYGWIN__)
 typedef unsigned short sexp_tag_t;
 typedef unsigned int sexp_uint_t;
 typedef int sexp_sint_t;
+#define SEXP_PRIdFIXNUM "d"
 #define sexp_heap_align(n) sexp_align(n, 5)
 #define sexp_heap_chunks(n) (sexp_heap_align(n)>>5)
+#elif PLAN9
+typedef uintptr sexp_tag_t;
+typedef unsigned int sexp_uint_t;
+typedef int sexp_sint_t;
+#define SEXP_PRIdFIXNUM "d"
+#define sexp_heap_align(n) sexp_align(n, 4)
+#define sexp_heap_chunks(n) (sexp_heap_align(n)>>4)
 #else
 typedef unsigned short sexp_tag_t;
 typedef unsigned int sexp_uint_t;
 typedef int sexp_sint_t;
+#define SEXP_PRIdFIXNUM "d"
 #define sexp_heap_align(n) sexp_align(n, 4)
 #define sexp_heap_chunks(n) (sexp_heap_align(n)>>4)
 #endif
@@ -251,7 +273,11 @@ typedef int sexp_sint_t;
 
 
 #ifdef SEXP_USE_INTTYPES
-# include <inttypes.h>
+#ifdef PLAN9
+#include <ape/stdint.h>
+#else
+#include <stdint.h>
+#endif
 # ifdef UINT8_MAX
 #  define SEXP_UINT8_DEFINED 1
 typedef uint8_t  sexp_uint8_t;
@@ -266,6 +292,13 @@ typedef int32_t sexp_int32_t;
 # include <ape/limits.h>
 # else
 # include <limits.h>
+# if SEXP_USE_UNIFORM_VECTOR_LITERALS
+# ifdef PLAN9
+# include <ape/stdint.h>
+# else
+# include <stdint.h>
+# endif
+# endif
 # endif
 # if UCHAR_MAX == 255
 #  define SEXP_UINT8_DEFINED 1
@@ -284,12 +317,12 @@ typedef long sexp_int32_t;
 typedef unsigned short sexp_uint32_t;
 typedef short sexp_int32_t;
 # endif
-#endif
+#endif  /* SEXP_USE_INTTYPES */
 
 #if defined(__APPLE__) || defined(_WIN64) || (defined(__CYGWIN__) && __SIZEOF_POINTER__ == 8)
-#define PRIoff "%lld"
+#define SEXP_PRIdOFF "lld"
 #else
-#define PRIoff "%ld"
+#define SEXP_PRIdOFF "ld"
 #endif
 
 #if SEXP_USE_LONG_PROCEDURE_ARGS
@@ -367,6 +400,7 @@ struct sexp_library_entry_t {   /* for static builds */
 };
 
 struct sexp_type_struct {
+  sexp name, cpl, slots, getters, setters, id, print, dl, finalize_name;
   sexp_tag_t tag;
   short field_base, field_eq_len_base, field_len_base, field_len_off;
   unsigned short field_len_scale;
@@ -374,14 +408,13 @@ struct sexp_type_struct {
   unsigned short size_scale;
   short weak_base, weak_len_base, weak_len_off, weak_len_scale, weak_len_extra;
   short depth;
-  sexp name, cpl, slots, getters, setters, id, print, dl, finalize_name;
   sexp_proc2 finalize;
 };
 
 struct sexp_opcode_struct {
-  unsigned char op_class, code, num_args, flags, inverse;
   sexp name, data, data2, proc, ret_type, arg1_type, arg2_type, arg3_type,
     argn_type, methods, dl;
+  unsigned char op_class, code, num_args, flags, inverse;
   sexp_proc1 func;
 };
 
@@ -390,6 +423,17 @@ struct sexp_core_form_struct {
   sexp name;
 };
 
+struct sexp_mark_stack_ptr_t {
+  sexp *start, *end;
+  struct sexp_mark_stack_ptr_t *prev; /* TODO: remove for allocations on stack */
+};
+
+/* Note this must be kept in sync with the _sexp_type_specs type            */
+/* registry in sexp.c.  The structure of a sexp type is:                    */
+/*   [ HEADER [[EQ_FIELDS... ] GC_FIELDS...] [WEAK_FIELDS...] [OTHER...] ]  */
+/* Thus all sexp's must be contiguous and align at the start of the type.   */
+/* This is used by the gc, equal? and slot-ref (although only the latter    */
+/* expects the alignment at the start of the type). */
 struct sexp_struct {
   sexp_tag_t tag;
   char markedp;
@@ -422,12 +466,23 @@ struct sexp_struct {
       char data SEXP_FLEXIBLE_ARRAY;
     } bytes;
     struct {
+      sexp bytes;
+      unsigned char element_type;
+      sexp_sint_t length;
+    } uvector;
+    struct {
 #if SEXP_USE_PACKED_STRINGS
+#if SEXP_USE_STRING_INDEX_TABLE
+      sexp charlens;
+#endif
       sexp_uint_t length;
       char data SEXP_FLEXIBLE_ARRAY;
 #else
-      sexp_uint_t offset, length;
       sexp bytes;
+#if SEXP_USE_STRING_INDEX_TABLE
+      sexp charlens;
+#endif
+      sexp_uint_t offset, length;
 #endif
     } string;
     struct {
@@ -435,22 +490,22 @@ struct sexp_struct {
       char data SEXP_FLEXIBLE_ARRAY;
     } symbol;
     struct {
+      sexp name;
+      sexp cookie;
+      sexp fd;
       FILE *stream;
       char *buf;
       char openp, bidirp, binaryp, shutdownp, no_closep, sourcep,
         blockedp, fold_casep;
       sexp_uint_t offset, line, flags;
       size_t size;
-      sexp name;
-      sexp cookie;
-      sexp fd;
     } port;
     struct {
       char openp, no_closep;
       sexp_sint_t fd, count;
     } fileno;
     struct {
-      sexp kind, message, irritants, procedure, source;
+      sexp kind, message, irritants, procedure, source, stack_trace;
     } exception;
     struct {
       signed char sign;
@@ -464,33 +519,33 @@ struct sexp_struct {
       sexp real, imag;
     } complex;
     struct {
+      sexp parent;
       sexp_uint_t length;
       void *value;
-      sexp parent;
       char body SEXP_FLEXIBLE_ARRAY;
     } cpointer;
     /* runtime types */
     struct {
       sexp parent, lambda, bindings;
-#if SEXP_USE_RENAME_BINDINGS
+#if SEXP_USE_STABLE_ABI || SEXP_USE_RENAME_BINDINGS
       sexp renames;
 #endif
     } env;
     struct {
-      sexp_uint_t length, max_depth;
       sexp name, literals, source;
+      sexp_uint_t length, max_depth;
       unsigned char data SEXP_FLEXIBLE_ARRAY;
     } bytecode;
     struct {
+      sexp bc, vars;
       char flags;
       sexp_proc_num_args_t num_args;
-      sexp bc, vars;
     } procedure;
     struct {
-      sexp proc, env, source;
+      sexp proc, env, source, aux;
     } macro;
     struct {
-      sexp env, free_vars, expr;
+      sexp env, free_vars, expr, rename;
     } synclo;
     struct {
       sexp file;
@@ -526,31 +581,41 @@ struct sexp_struct {
       sexp data SEXP_FLEXIBLE_ARRAY;
     } stack;
     struct {
+      sexp stack, env, parent, child,
+        globals, dk, params, proc, name, specific, event, result;
+#if SEXP_USE_STABLE_ABI || SEXP_USE_DL
+      sexp dl;
+#endif
       sexp_heap heap;
+      struct sexp_mark_stack_ptr_t mark_stack[SEXP_MARK_STACK_COUNT];
+      struct sexp_mark_stack_ptr_t *mark_stack_ptr;
       struct sexp_gc_var_t *saves;
 #if SEXP_USE_GREEN_THREADS
       sexp_sint_t refuel;
       unsigned char* ip;
       struct timeval tval;
 #endif
-      char tailp, tracep, timeoutp, waitp, errorp;
+      char tailp, tracep, timeoutp, waitp, errorp, interruptp;
       sexp_uint_t last_fp;
+      sexp_uint_t gc_count;
 #if SEXP_USE_TIME_GC
-      sexp_uint_t gc_count, gc_usecs;
+      sexp_uint_t gc_usecs;
 #endif
-      sexp stack, env, parent, child,
-        globals, dk, params, proc, name, specific, event, result;
-#if SEXP_USE_DL
-      sexp dl;
+#if SEXP_USE_TRACK_ALLOC_TIMES
+      sexp_uint_t alloc_count, alloc_usecs;
+      double alloc_usecs_sq;
+#endif
+#if SEXP_USE_TRACK_ALLOC_SIZES
+      sexp_uint_t alloc_histogram[SEXP_ALLOC_HISTOGRAM_BUCKETS];
 #endif
     } context;
-#if SEXP_USE_AUTO_FORCE
+#if SEXP_USE_STABLE_ABI || SEXP_USE_AUTO_FORCE
     struct {
-      int donep;
       sexp value;
+      int donep;
     } promise;
 #endif
-#if SEXP_USE_WEAK_REFERENCES
+#if SEXP_USE_STABLE_ABI || SEXP_USE_WEAK_REFERENCES
     struct {
       sexp key, value;
     } ephemeron;
@@ -571,9 +636,10 @@ struct sexp_struct {
 #define SEXP_RAWDOT SEXP_MAKE_IMMEDIATE(7) /* internal use */
 #define SEXP_STRING_OPORT SEXP_MAKE_IMMEDIATE(8)  /* internal use */
 #define SEXP_TRAMPOLINE   SEXP_MAKE_IMMEDIATE(9)  /* internal use */
-#define SEXP_ABI_ERROR    SEXP_MAKE_IMMEDIATE(10) /* internal use */
+#define SEXP_UNCAUGHT     SEXP_MAKE_IMMEDIATE(10) /* internal use */
+#define SEXP_ABI_ERROR    SEXP_MAKE_IMMEDIATE(11) /* internal use */
 #if SEXP_USE_OBJECT_BRACE_LITERALS
-#define SEXP_CLOSE_BRACE  SEXP_MAKE_IMMEDIATE(11) /* internal use */
+#define SEXP_CLOSE_BRACE  SEXP_MAKE_IMMEDIATE(12) /* internal use */
 #endif
 
 #if SEXP_USE_LIMITED_MALLOC
@@ -737,6 +803,7 @@ union sexp_flonum_conv {
 SEXP_API sexp sexp_flonum_predicate (sexp ctx, sexp x);
 #if SEXP_64_BIT
 SEXP_API float sexp_flonum_value (sexp x);
+#define sexp_flonum_value_set(f, x) (f = sexp_make_flonum(NULL, x))
 #define sexp_flonum_bits(f) ((char*)&f)
 SEXP_API sexp sexp_make_flonum(sexp ctx, float f);
 #else
@@ -746,6 +813,7 @@ SEXP_API sexp sexp_make_flonum(sexp ctx, float f);
 #else
 #define sexp_flonump(x)      (sexp_check_tag(x, SEXP_FLONUM))
 #define sexp_flonum_value(f) ((f)->value.flonum)
+#define sexp_flonum_value_set(f, x) ((f)->value.flonum = x)
 #define sexp_flonum_bits(f) ((f)->value.flonum_bits)
 SEXP_API sexp sexp_make_flonum(sexp ctx, double f);
 #endif
@@ -800,6 +868,38 @@ SEXP_API sexp sexp_make_flonum(sexp ctx, double f);
 #define sexp_promisep(x)    (sexp_check_tag(x, SEXP_PROMISE))
 #define sexp_ephemeronp(x)  (sexp_check_tag(x, SEXP_EPHEMERON))
 
+#if SEXP_USE_UNIFORM_VECTOR_LITERALS
+#define sexp_uvectorp(x)    (sexp_check_tag(x, SEXP_UNIFORM_VECTOR))
+#define sexp_u1vectorp(x)   (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_U1)
+#define sexp_u8vectorp(x)   (sexp_bytesp(x))
+#define sexp_s8vectorp(x)   (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_S8)
+#define sexp_u16vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_U16)
+#define sexp_s16vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_S16)
+#define sexp_u32vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_U32)
+#define sexp_s32vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_S32)
+#define sexp_u64vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_U64)
+#define sexp_s64vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_S64)
+#define sexp_f32vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_F32)
+#define sexp_f64vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_F64)
+#define sexp_c64vectorp(x)  (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_C64)
+#define sexp_c128vectorp(x) (sexp_uvectorp(x) && sexp_uvector_type(x)==SEXP_C128)
+#else
+#define sexp_uvectorp(x)    (sexp_vectorp(x))
+#define sexp_u1vectorp(x)   (sexp_vectorp(x))
+#define sexp_u8vectorp(x)   (sexp_bytesp(x))
+#define sexp_s8vectorp(x)   (sexp_vectorp(x))
+#define sexp_u16vectorp(x)  (sexp_vectorp(x))
+#define sexp_s16vectorp(x)  (sexp_vectorp(x))
+#define sexp_u32vectorp(x)  (sexp_vectorp(x))
+#define sexp_s32vectorp(x)  (sexp_vectorp(x))
+#define sexp_u64vectorp(x)  (sexp_vectorp(x))
+#define sexp_s64vectorp(x)  (sexp_vectorp(x))
+#define sexp_f32vectorp(x)  (sexp_vectorp(x))
+#define sexp_f64vectorp(x)  (sexp_vectorp(x))
+#define sexp_c64vectorp(x)  (sexp_vectorp(x))
+#define sexp_c128vectorp(x) (sexp_vectorp(x))
+#endif
+
 #define sexp_applicablep(x) (sexp_procedurep(x) || sexp_opcodep(x))
 
 #if SEXP_USE_HUFF_SYMS
@@ -891,8 +991,15 @@ SEXP_API int sexp_idp(sexp x);
 #endif
 
 #if SEXP_USE_BIGNUMS
+SEXP_API sexp sexp_make_integer_from_lsint(sexp ctx, sexp_lsint_t x);
+SEXP_API sexp sexp_make_unsigned_integer_from_luint(sexp ctx, sexp_luint_t x);
+#if SEXP_USE_CUSTOM_LONG_LONGS
+SEXP_API sexp sexp_make_integer(sexp ctx, long long x);
+SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, unsigned long long x);
+#else
 SEXP_API sexp sexp_make_integer(sexp ctx, sexp_lsint_t x);
 SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
+#endif
 #define sexp_exact_integerp(x) (sexp_fixnump(x) || sexp_bignump(x))
 #else
 #define sexp_make_integer(ctx, x) sexp_make_fixnum(x)
@@ -922,8 +1029,12 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
 
 #if SEXP_USE_COMPLEX
 #define sexp_numberp(x) (sexp_realp(x) || sexp_complexp(x))
+#define sexp_real_part(x) (sexp_complexp(x) ? sexp_complex_real(x) : x)
+#define sexp_imag_part(x) (sexp_complexp(x) ? sexp_complex_imag(x) : SEXP_ZERO)
 #else
 #define sexp_numberp(x) (sexp_realp(x))
+#define sexp_real_part(x) (x)
+#define sexp_imag_part(x) SEXP_ZERO
 #endif
 
 #define sexp_exact_negativep(x) (sexp_fixnump(x) ? (sexp_unbox_fixnum(x) < 0) \
@@ -968,12 +1079,24 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
     sexp_negate_exact(x)
 
 #if SEXP_USE_FLONUMS || SEXP_USE_BIGNUMS
-#define sexp_uint_value(x) ((sexp_uint_t)(sexp_fixnump(x) ? sexp_unbox_fixnum(x) : sexp_bignump(x) ? sexp_bignum_data(x)[0] : 0))
-#define sexp_sint_value(x) ((sexp_sint_t)(sexp_fixnump(x) ? sexp_unbox_fixnum(x) : sexp_bignump(x) ? sexp_bignum_sign(x)*sexp_bignum_data(x)[0] : 0))
+
+#if SEXP_64_BIT
+#define sexp_bignum_to_sint(x) (sexp_bignum_sign(x)*sexp_bignum_data(x)[0])
+#define sexp_bignum_to_uint(x) (sexp_bignum_data(x)[0])
 #else
+SEXP_API long long sexp_bignum_to_sint(sexp x);
+SEXP_API unsigned long long sexp_bignum_to_uint(sexp x);
+#endif
+
+#define sexp_uint_value(x) ((unsigned long long)(sexp_fixnump(x) ? sexp_unbox_fixnum(x) : sexp_bignump(x) ? sexp_bignum_to_uint(x) : 0))
+#define sexp_sint_value(x) ((long long)(sexp_fixnump(x) ? sexp_unbox_fixnum(x) : sexp_bignump(x) ? sexp_bignum_to_sint(x) : 0))
+
+#else
+
 #define sexp_uint_value(x) ((sexp_uint_t)sexp_unbox_fixnum(x))
 #define sexp_sint_value(x) ((sexp_sint_t)sexp_unbox_fixnum(x))
-#endif
+
+#endif	/* SEXP_USE_FLONUMS || SEXP_USE_BIGNUMS */
 
 #define sexp_shift_epoch(x) ((x)-SEXP_EPOCH_OFFSET)
 #define sexp_unshift_epoch(x) ((x)+SEXP_EPOCH_OFFSET)
@@ -1028,7 +1151,42 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
 #define sexp_bytes_data(x)    (sexp_field(x, bytes, SEXP_BYTES, data))
 #define sexp_bytes_maybe_null_data(x) (sexp_not(x) ? NULL : sexp_bytes_data(x))
 
-#define sexp_string_size(x)   (sexp_field(x, string, SEXP_STRING, length))
+static const unsigned char sexp_uvector_sizes[] = {
+  0, 1, 8, 8, 16, 16, 32, 32, 64, 64, 32, 64, 64, 128};
+static const unsigned char sexp_uvector_chars[] = "#ususususuffcc";
+
+enum sexp_uniform_vector_type {
+  SEXP_NOT_A_UNIFORM_TYPE,
+  SEXP_U1,
+  SEXP_S8,
+  SEXP_U8,
+  SEXP_S16,
+  SEXP_U16,
+  SEXP_S32,
+  SEXP_U32,
+  SEXP_S64,
+  SEXP_U64,
+  SEXP_F32,
+  SEXP_F64,
+  SEXP_C64,
+  SEXP_C128
+};
+
+#define sexp_uvector_freep(x) (sexp_freep(x))
+#define sexp_uvector_element_size(uvt) (sexp_uvector_sizes[uvt])
+#define sexp_uvector_prefix(uvt) (sexp_uvector_chars[uvt])
+
+#define sexp_uvector_length(x) (sexp_field(x, uvector, SEXP_UNIFORM_VECTOR, length))
+#define sexp_uvector_type(x)   (sexp_field(x, uvector, SEXP_UNIFORM_VECTOR, element_type))
+#define sexp_uvector_data(x) sexp_bytes_data(sexp_uvector_bytes(x))
+#define sexp_uvector_maybe_null_data(x) (sexp_not(x) ? NULL : sexp_uvector_data(x))
+#define sexp_uvector_bytes(x)  (sexp_field(x, uvector, SEXP_UNIFORM_VECTOR, bytes))
+
+#define sexp_bit_ref(u1v, i)    (((sexp_uvector_data(u1v)[i/8])>>(i%8))&1)
+#define sexp_bit_set(u1v, i, x) (x ? (sexp_uvector_data(u1v)[i/8]|=(1<<(i%8))) : (sexp_uvector_data(u1v)[i/8]&=~(1<<(i%8))))
+
+#define sexp_string_size(x)     (sexp_field(x, string, SEXP_STRING, length))
+#define sexp_string_charlens(x) (sexp_field(x, string, SEXP_STRING, charlens))
 #if SEXP_USE_PACKED_STRINGS
 #define sexp_string_data(x)   (sexp_field(x, string, SEXP_STRING, data))
 #define sexp_string_bytes(x)  (x)
@@ -1086,6 +1244,7 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
 #define sexp_exception_irritants(x) (sexp_field(x, exception, SEXP_EXCEPTION, irritants))
 #define sexp_exception_procedure(x) (sexp_field(x, exception, SEXP_EXCEPTION, procedure))
 #define sexp_exception_source(x)    (sexp_field(x, exception, SEXP_EXCEPTION, source))
+#define sexp_exception_stack_trace(x) (sexp_field(x, exception, SEXP_EXCEPTION, stack_trace))
 
 #define sexp_trampolinep(x) (sexp_exceptionp(x) && sexp_exception_kind(x) == SEXP_TRAMPOLINE)
 #define sexp_trampoline_procedure(x) sexp_exception_procedure(x)
@@ -1119,10 +1278,12 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
 #define sexp_macro_proc(x)        (sexp_field(x, macro, SEXP_MACRO, proc))
 #define sexp_macro_env(x)         (sexp_field(x, macro, SEXP_MACRO, env))
 #define sexp_macro_source(x)      (sexp_field(x, macro, SEXP_MACRO, source))
+#define sexp_macro_aux(x)         (sexp_field(x, macro, SEXP_MACRO, aux))
 
 #define sexp_synclo_env(x)        (sexp_field(x, synclo, SEXP_SYNCLO, env))
 #define sexp_synclo_free_vars(x)  (sexp_field(x, synclo, SEXP_SYNCLO, free_vars))
 #define sexp_synclo_expr(x)       (sexp_field(x, synclo, SEXP_SYNCLO, expr))
+#define sexp_synclo_rename(x)     (sexp_field(x, synclo, SEXP_SYNCLO, rename))
 
 #define sexp_core_code(x)         (sexp_field(x, core, SEXP_CORE, code))
 #define sexp_core_name(x)         (sexp_field(x, core, SEXP_CORE, name))
@@ -1204,6 +1365,8 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
 #define sexp_context_stack(x)    (sexp_field(x, context, SEXP_CONTEXT, stack))
 #define sexp_context_parent(x)   (sexp_field(x, context, SEXP_CONTEXT, parent))
 #define sexp_context_child(x)    (sexp_field(x, context, SEXP_CONTEXT, child))
+#define sexp_context_mark_stack(x)     (sexp_field(x, context, SEXP_CONTEXT, mark_stack))
+#define sexp_context_mark_stack_ptr(x) (sexp_field(x, context, SEXP_CONTEXT, mark_stack_ptr))
 #define sexp_context_saves(x)    (sexp_field(x, context, SEXP_CONTEXT, saves))
 #define sexp_context_tailp(x)    (sexp_field(x, context, SEXP_CONTEXT, tailp))
 #define sexp_context_tracep(x)   (sexp_field(x, context, SEXP_CONTEXT, tracep))
@@ -1211,13 +1374,20 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
 #define sexp_context_dk(x)       (sexp_field(x, context, SEXP_CONTEXT, dk))
 #define sexp_context_params(x)   (sexp_field(x, context, SEXP_CONTEXT, params))
 #define sexp_context_last_fp(x)  (sexp_field(x, context, SEXP_CONTEXT, last_fp))
-#if SEXP_USE_TIME_GC
 #define sexp_context_gc_count(x) (sexp_field(x, context, SEXP_CONTEXT, gc_count))
+#if SEXP_USE_TIME_GC
 #define sexp_context_gc_usecs(x) (sexp_field(x, context, SEXP_CONTEXT, gc_usecs))
 #else
-#define sexp_context_gc_count(x) 0
 #define sexp_context_gc_usecs(x) 0
 #endif
+#if SEXP_USE_TRACK_ALLOC_TIMES
+#define sexp_context_alloc_count(x) (sexp_field(x, context, SEXP_CONTEXT, alloc_count))
+#define sexp_context_alloc_usecs(x) (sexp_field(x, context, SEXP_CONTEXT, alloc_usecs))
+#define sexp_context_alloc_usecs_sq(x) (sexp_field(x, context, SEXP_CONTEXT, alloc_usecs_sq))
+#endif
+#if SEXP_USE_TRACK_ALLOC_SIZES
+#define sexp_context_alloc_histogram(x) (sexp_field(x, context, SEXP_CONTEXT, alloc_histogram))
+#endif
 #define sexp_context_refuel(x)   (sexp_field(x, context, SEXP_CONTEXT, refuel))
 #define sexp_context_ip(x)       (sexp_field(x, context, SEXP_CONTEXT, ip))
 #define sexp_context_proc(x)     (sexp_field(x, context, SEXP_CONTEXT, proc))
@@ -1231,6 +1401,7 @@ SEXP_API sexp sexp_make_unsigned_integer(sexp ctx, sexp_luint_t x);
 
 #define sexp_context_result(x)   (sexp_field(x, context, SEXP_CONTEXT, result))
 #define sexp_context_errorp(x)   (sexp_field(x, context, SEXP_CONTEXT, errorp))
+#define sexp_context_interruptp(x) (sexp_field(x, context, SEXP_CONTEXT, interruptp))
 
 /* during compilation, sexp_context_specific is set to a vector */
 /* containing the following elements: */
@@ -1357,15 +1528,17 @@ SEXP_API sexp sexp_symbol_table[SEXP_SYMBOL_TABLE_SIZE];
 /****************************** utilities *****************************/
 
 enum sexp_context_globals {
-#if ! SEXP_USE_GLOBAL_SYMBOLS
+#if SEXP_USE_STABLE_ABI || ! SEXP_USE_GLOBAL_SYMBOLS
   SEXP_G_SYMBOLS,
 #endif
+  SEXP_G_ENDIANNESS,
   SEXP_G_TYPES,
   SEXP_G_FEATURES,
   SEXP_G_NUM_TYPES,
   SEXP_G_OOM_ERROR,             /* out of memory exception object */
   SEXP_G_OOS_ERROR,             /* out of stack exception object */
   SEXP_G_ABI_ERROR,             /* incompatible ABI loading library */
+  SEXP_G_INTERRUPT_ERROR,       /* C-c in the repl */
   SEXP_G_OPTIMIZATIONS,
   SEXP_G_SIGNAL_HANDLERS,
   SEXP_G_META_ENV,
@@ -1374,6 +1547,10 @@ enum sexp_context_globals {
   SEXP_G_QUASIQUOTE_SYMBOL,
   SEXP_G_UNQUOTE_SYMBOL,
   SEXP_G_UNQUOTE_SPLICING_SYMBOL,
+  SEXP_G_SYNTAX_SYMBOL,
+  SEXP_G_QUASISYNTAX_SYMBOL,
+  SEXP_G_UNSYNTAX_SYMBOL,
+  SEXP_G_UNSYNTAX_SPLICING_SYMBOL,
   SEXP_G_EMPTY_VECTOR,
   SEXP_G_CUR_IN_SYMBOL,
   SEXP_G_CUR_OUT_SYMBOL,
@@ -1386,18 +1563,18 @@ enum sexp_context_globals {
   SEXP_G_RANDOM_SOURCE,
   SEXP_G_STRICT_P,
   SEXP_G_NO_TAIL_CALLS_P,
-#if SEXP_USE_FOLD_CASE_SYMS
+#if SEXP_USE_STABLE_ABI || SEXP_USE_FOLD_CASE_SYMS
   SEXP_G_FOLD_CASE_P,
 #endif
-#if SEXP_USE_WEAK_REFERENCES
+#if SEXP_USE_STABLE_ABI || SEXP_USE_WEAK_REFERENCES
   SEXP_G_WEAK_OBJECTS_PRESENT,
   SEXP_G_FILE_DESCRIPTORS,
   SEXP_G_NUM_FILE_DESCRIPTORS,
 #endif
-#if ! SEXP_USE_BOEHM
+#if SEXP_USE_STABLE_ABI || ! SEXP_USE_BOEHM
   SEXP_G_PRESERVATIVES,
 #endif
-#if SEXP_USE_GREEN_THREADS
+#if SEXP_USE_STABLE_ABI || SEXP_USE_GREEN_THREADS
   SEXP_G_IO_BLOCK_ERROR,
   SEXP_G_IO_BLOCK_ONCE_ERROR,
   SEXP_G_THREAD_TERMINATE_ERROR,
@@ -1499,6 +1676,7 @@ SEXP_API sexp sexp_alloc_tagged_aux(sexp ctx, size_t size, sexp_uint_t tag sexp_
 SEXP_API sexp sexp_make_context(sexp ctx, size_t size, size_t max_size);
 SEXP_API sexp sexp_cons_op(sexp ctx, sexp self, sexp_sint_t n, sexp head, sexp tail);
 SEXP_API sexp sexp_list2(sexp ctx, sexp a, sexp b);
+SEXP_API sexp sexp_list3(sexp ctx, sexp a, sexp b, sexp c);
 SEXP_API sexp sexp_equalp_bound (sexp ctx, sexp self, sexp_sint_t n, sexp a, sexp b, sexp depth, sexp bound);
 SEXP_API sexp sexp_equalp_op (sexp ctx, sexp self, sexp_sint_t n, sexp a, sexp b);
 SEXP_API sexp sexp_listp_op(sexp ctx, sexp self, sexp_sint_t n, sexp obj);
@@ -1512,6 +1690,7 @@ SEXP_API sexp sexp_length_op(sexp ctx, sexp self, sexp_sint_t n, sexp ls);
 SEXP_API sexp sexp_c_string(sexp ctx, const char *str, sexp_sint_t slen);
 SEXP_API sexp sexp_make_ephemeron_op(sexp ctx, sexp self, sexp_sint_t n, sexp key, sexp value);
 SEXP_API sexp sexp_make_bytes_op(sexp ctx, sexp self, sexp_sint_t n, sexp len, sexp i);
+SEXP_API sexp sexp_make_uvector_op(sexp ctx, sexp self, sexp_sint_t n, sexp elt_type, sexp len);
 SEXP_API sexp sexp_make_string_op(sexp ctx, sexp self, sexp_sint_t n, sexp len, sexp ch);
 SEXP_API sexp sexp_substring_op (sexp ctx, sexp self, sexp_sint_t n, sexp str, sexp start, sexp end);
 SEXP_API sexp sexp_subbytes_op (sexp ctx, sexp self, sexp_sint_t n, sexp str, sexp start, sexp end);
@@ -1523,6 +1702,7 @@ SEXP_API sexp sexp_string_to_number_op (sexp ctx, sexp self, sexp_sint_t n, sexp
 SEXP_API sexp sexp_flonump_op (sexp ctx, sexp self, sexp_sint_t n, sexp x);
 SEXP_API sexp sexp_make_vector_op (sexp ctx, sexp self, sexp_sint_t n, sexp len, sexp dflt);
 SEXP_API sexp sexp_list_to_vector_op (sexp ctx, sexp self, sexp_sint_t n, sexp ls);
+SEXP_API sexp sexp_list_to_uvector_op (sexp ctx, sexp self, sexp_sint_t n, sexp etype, sexp ls);
 SEXP_API sexp sexp_make_cpointer (sexp ctx, sexp_uint_t type_id, void* value, sexp parent, int freep);
 SEXP_API int sexp_is_separator(int c);
 SEXP_API sexp sexp_write_op (sexp ctx, sexp self, sexp_sint_t n, sexp obj, sexp out);
@@ -1532,7 +1712,7 @@ SEXP_API sexp sexp_read_symbol (sexp ctx, sexp in, int init, int internp);
 SEXP_API sexp sexp_read_number (sexp ctx, sexp in, int base, int exactp);
 #if SEXP_USE_BIGNUMS
 SEXP_API sexp sexp_read_bignum (sexp ctx, sexp in, sexp_uint_t init,
-				signed char sign, sexp_uint_t base);
+                                signed char sign, sexp_uint_t base);
 SEXP_API sexp sexp_write_bignum (sexp ctx, sexp a, sexp out, sexp_uint_t base);
 #endif
 SEXP_API sexp sexp_read_float_tail(sexp ctx, sexp in, double whole, int negp);
@@ -1570,15 +1750,18 @@ SEXP_API sexp sexp_open_output_string_op (sexp ctx, sexp self, sexp_sint_t n);
 SEXP_API sexp sexp_get_output_string_op (sexp ctx, sexp self, sexp_sint_t n, sexp port);
 SEXP_API sexp sexp_make_exception (sexp ctx, sexp kind, sexp message, sexp irritants, sexp procedure, sexp source);
 SEXP_API sexp sexp_user_exception (sexp ctx, sexp self, const char *msg, sexp x);
+SEXP_API sexp sexp_user_exception_ls (sexp ctx, sexp self, const char *msg, int n, ...);
 SEXP_API sexp sexp_file_exception (sexp ctx, sexp self, const char *msg, sexp x);
 SEXP_API sexp sexp_type_exception (sexp ctx, sexp self, sexp_uint_t type_id, sexp x);
 SEXP_API sexp sexp_xtype_exception (sexp ctx, sexp self, const char *msg, sexp x);
 SEXP_API sexp sexp_range_exception (sexp ctx, sexp obj, sexp start, sexp end);
 SEXP_API sexp sexp_print_exception_op (sexp ctx, sexp self, sexp_sint_t n, sexp exn, sexp out);
 SEXP_API sexp sexp_stack_trace_op (sexp ctx, sexp self, sexp_sint_t n, sexp out);
+SEXP_API sexp sexp_print_exception_stack_trace_op (sexp ctx, sexp self, sexp_sint_t n, sexp exn, sexp out);
 SEXP_API sexp sexp_apply (sexp context, sexp proc, sexp args);
 SEXP_API sexp sexp_apply1 (sexp ctx, sexp f, sexp x);
 SEXP_API sexp sexp_apply2 (sexp ctx, sexp f, sexp x, sexp y);
+SEXP_API sexp sexp_apply3 (sexp ctx, sexp f, sexp x, sexp y, sexp z);
 SEXP_API sexp sexp_apply_no_err_handler (sexp ctx, sexp proc, sexp args);
 SEXP_API sexp sexp_make_trampoline (sexp ctx, sexp proc, sexp args);
 SEXP_API sexp sexp_make_foreign (sexp ctx, const char *name, int num_args, int flags, const char *fname, sexp_proc1 f, sexp data);
@@ -1618,6 +1801,12 @@ SEXP_API int sexp_write_utf8_char (sexp ctx, int c, sexp out);
 #define sexp_substring_cursor(ctx, s, i, j) sexp_substring_op(ctx, NULL, 3, s, i, j)
 #endif
 
+#if SEXP_USE_STRING_INDEX_TABLE
+SEXP_API void sexp_update_string_index_lookup(sexp ctx, sexp s);
+#else
+#define sexp_update_string_index_lookup(ctx, s)
+#endif
+
 #if SEXP_USE_GREEN_THREADS
 SEXP_API int sexp_maybe_block_port (sexp ctx, sexp in, int forcep);
 SEXP_API void sexp_maybe_unblock_port (sexp ctx, sexp in);
@@ -1653,9 +1842,13 @@ SEXP_API sexp sexp_finalize (sexp ctx);
 
 #if SEXP_USE_GLOBAL_HEAP
 #define sexp_free_heap(heap)
+#define sexp_debug_heap_stats(heap)
 #define sexp_destroy_context(ctx) SEXP_TRUE
 #else
 SEXP_API void sexp_free_heap (sexp_heap heap);
+SEXP_API void sexp_debug_heap_stats (sexp_heap heap);
+SEXP_API void sexp_debug_alloc_times(sexp ctx);
+SEXP_API void sexp_debug_alloc_sizes(sexp ctx);
 SEXP_API sexp sexp_destroy_context (sexp ctx);
 SEXP_API sexp sexp_copy_context (sexp ctx, sexp dst, sexp flags);
 #endif
@@ -1699,6 +1892,7 @@ SEXP_API int sexp_poll_port(sexp ctx, sexp port, int inputp);
 #define sexp_read(ctx, in) sexp_read_op(ctx, NULL, 1, in)
 #define sexp_write(ctx, obj, out) sexp_write_op(ctx, NULL, 2, obj, out)
 #define sexp_print_exception(ctx, e, out) sexp_print_exception_op(ctx, NULL, 2, e, out)
+#define sexp_print_exception_stack_trace(ctx, e, out) sexp_print_exception_stack_trace_op(ctx, NULL, 2, e, out)
 #define sexp_flush_output(ctx, out) sexp_flush_output_op(ctx, NULL, 1, out)
 #define sexp_equalp(ctx, a, b) sexp_equalp_op(ctx, NULL, 2, a, b)
 #define sexp_listp(ctx, x) sexp_listp_op(ctx, NULL, 1, x)
@@ -1711,12 +1905,20 @@ SEXP_API int sexp_poll_port(sexp ctx, sexp port, int inputp);
 #define sexp_append2(ctx, a, b) sexp_append2_op(ctx, NULL, 2, a, b)
 #define sexp_make_vector(ctx, a, b) sexp_make_vector_op(ctx, NULL, 2, a, b)
 #define sexp_list_to_vector(ctx, x) sexp_list_to_vector_op(ctx, NULL, 1, x)
+#define sexp_list_to_uvector(ctx, etype, ls) sexp_list_to_uvector_op(ctx, NULL, 2, etype, ls)
 #define sexp_exception_type(ctx, x) sexp_exception_type_op(ctx, NULL, 1, x)
 #define sexp_string_to_symbol(ctx, s) sexp_string_to_symbol_op(ctx, NULL, 1, s)
 #define sexp_string_to_number(ctx, s, b) sexp_string_to_number_op(ctx, NULL, 2, s, b)
 #define sexp_symbol_to_string(ctx, s) sexp_symbol_to_string_op(ctx, NULL, 1, s)
 #define sexp_make_ephemeron(ctx, k, v) sexp_make_ephemeron_op(ctx, NULL, 2, k, v)
 #define sexp_make_bytes(ctx, l, i) sexp_make_bytes_op(ctx, NULL, 2, l, i)
+#if SEXP_USE_UNIFORM_VECTOR_LITERALS
+#define sexp_make_uvector(ctx, et, l) sexp_make_uvector_op(ctx, NULL, 2, et, l)
+#else
+#define sexp_make_uvector(ctx, et, l) sexp_make_vector(ctx, l, SEXP_ZERO)
+#define sexp_write_uvector NULL
+#define sexp_finalize_uvector NULL
+#endif
 #define sexp_make_string(ctx, l, c) sexp_make_string_op(ctx, NULL, 2, l, c)
 #define sexp_subbytes(ctx, a, b, c) sexp_subbytes_op(ctx, NULL, 3, a, b, c)
 #define sexp_string_concatenate(ctx, ls, s) sexp_string_concatenate_op(ctx, NULL, 2, ls, s)
