@@ -1,20 +1,28 @@
 ;; http-server.scm -- combinator-based http server
-;; Copyright (c) 2013-2015 Alex Shinn.  All rights reserved.
+;; Copyright (c) 2013-2019 Alex Shinn.  All rights reserved.
 ;; BSD-style license: http://synthcode.com/license.txt
 
 ;;> Runs an http server listening at the given address, with the given
 ;;> servlet.
 ;;>
-;;> An servlet is a procedure which takes four arguments: a 
+;;> A servlet is a procedure which takes four arguments: a 
 ;;> \scheme{(chibi config)} config object, an \scheme{Http-Request} record,
 ;;> which contains the I/O ports and parsed request and headers;
 ;;> a \scheme{next} procedure to call the next available servlet if any,
 ;;> and a \scheme{restart} procedure to restart the servlets with a new
 ;;> request.
 ;;>
+;;> The default config parameters are:
+;;>
+;;> \itemlist[
+;;> \item{\scheme{port}: the port to listen on, default 8000}
+;;> \item{\scheme{doc-root}: the directory to serve files from, default the directory of the config file, or "." if no config}
+;;> \item{\scheme{index-regexp}: an SRE matching index files to serve in place of a directory listing when browsing directories, default "index.html"}
+;;> ]
+;;>
 ;;> A simple page view counter could be run as:
 ;;>
-;;> \example{
+;;> \scheme{
 ;;> (let ((count 0))
 ;;>   (run-http-server
 ;;>    8000
@@ -32,16 +40,28 @@
         (cond
          ((= 2 (length ls))
           (let ((request
-                 (make-request command (car ls) (cadr ls) in out sock addr)))
-            (log-info `(request: ,command ,(car ls) ,(cadr ls)
-                                 ,(request-headers request)))
-            (protect (exn
+                 (protect
+                     (exn
                       (else
-                       (log-error "internal error: " exn)
-                       (print-stack-trace exn)
-                       (servlet-respond request 500 "Internal server error")))
-              (let restart ((request request))
-                (servlet cfg request servlet-bad-request restart)))))
+                       ;; error parsing headers, can't use servlet-respond
+                       (log-error "request error: " exn ls
+                                  (sockaddr-name (address-info-address addr)))
+                       (servlet-write-status out 500 "Internal server error")
+                       (mime-write-headers `((Status . "500")) out)
+                       (display "\r\n" out)
+                       #f))
+                   (make-request command (car ls) (cadr ls) in out sock addr))))
+            (cond
+             (request
+              (log-info `(request: ,command ,(car ls) ,(cadr ls)
+                                   ,(request-headers request)))
+              (protect (exn
+                        (else
+                         (log-error "internal error: " exn)
+                         (print-stack-trace exn)
+                         (servlet-respond request 500 "Internal server error")))
+                (let restart ((request request))
+                  (servlet cfg request servlet-bad-request restart)))))))
          (else
           (let ((request (make-request command #f #f in out sock addr)))
             (servlet-respond request 400 "bad request")))))))))
@@ -82,17 +102,20 @@
         (next cfg request))))
 
 ;; Generate a simple page listing the linked files in a directory.
-(define (send-directory path out)
-  (display "<html><body bgcolor=white><pre>\n" out)
-  (for-each
-   (lambda (file)
-     (display "<a href=\"/" out)
-     (display (path-normalize (make-path path file)) out)
-     (display "\">" out)
-     (display file out)
-     (display "</a>\n" out))
-   (sort (directory-files path)))
-  (display "</pre></body></html>\n" out))
+(define (send-directory path out . o)
+  (let ((base-dir (if (and (pair? o) (car o))
+                      (path-relative-to path (car o))
+                      path)))
+    (display "<html><body bgcolor=white><pre>\n" out)
+    (for-each
+     (lambda (file)
+       (display "<a href=\"/" out)
+       (display (path-normalize (make-path base-dir file)) out)
+       (display "\">" out)
+       (display file out)
+       (display "</a>\n" out))
+     (sort (directory-files path)))
+    (display "</pre></body></html>\n" out)))
 
 ;; TODO: If the index-rx is a short list of fixed strings, check
 ;; individually to avoid the full directory lookup.
@@ -101,7 +124,7 @@
        (any (lambda (f) (and (regexp-matches? index-rx f) (make-path dir f)))
             (directory-files dir))))
 
-(define (http-send-directory request path index-rx restart)
+(define (http-send-directory request path index-rx restart . o)
   (cond
    ((find-index-file path index-rx)
     => (lambda (index-file)
@@ -112,13 +135,19 @@
            (restart
             (request-with-uri request (uri-with-path uri path2))))))
    (else
-    (send-directory path (request-out request)))))
+    (servlet-respond request 200 "OK")
+    (apply send-directory path (request-out request) o))))
 
 (define (http-send-file request path)
   (cond
    ((file-exists? path)
-    (servlet-respond request 200 "OK")
-    (send-file path (request-out request)))
+    (let ((headers
+           (cond
+            ((mime-type-from-extension (path-extension path))
+             => (lambda (type) `((Content-Type . ,type))))
+            (else '()))))
+      (servlet-respond request 200 "OK" headers)
+      (send-file path (request-out request))))
    (else
     (servlet-respond request 404 "Not Found"))))
 
@@ -129,7 +158,7 @@
      (lambda (cfg request next restart)
        (let ((path (make-path dir (request-path request))))
          (if (file-directory? path)
-             (http-send-directory request path index-rx restart)
+             (http-send-directory request path index-rx restart dir)
              (http-send-file request path)))))))
 
 (define (http-procedure-servlet path proc)
@@ -336,7 +365,7 @@
   (case (request-method request)
     ((HEAD)
      (call-with-temp-file "get.out"
-       (lambda (temp-file out)
+       (lambda (temp-file out preserve)
          (let ((request2 (copy-request request)))
            (request-method-set! request2 'GET)
            (request-out-set! request2 out)
