@@ -3,31 +3,10 @@
 ;; This code was written by Alex Shinn in 2014 and placed in the
 ;; Public Domain.  All warranties are disclaimed.
 
-(define known-implementations
-  '((chibi "chibi-scheme" (chibi-scheme -V) "0.7.3")
-    (chicken "chicken" (csi -p "(chicken-version)") "4.9.0")
-    (cyclone "cyclone" (icyc -vn) "0.5.3")
-    (foment "foment")
-    (gauche "gosh" (gosh -E "print (gauche-version)" -E exit) "0.9.4")
-    (kawa "kawa" (kawa --version) "2.0")
-    (larceny "larceny" (larceny --version) "v0.98")
-    (sagittarius "sagittarius")))
-
-(define (impl->version impl cmd)
-  (let* ((lines (process->string-list cmd))
-         (line (and (pair? lines) (string-split (car lines)))))
-    (and (pair? line)
-         (if (and (pair? (cdr line))
-                  (let ((x (string-downcase (car line)))
-                        (name (symbol->string impl)))
-                    (or (equal? x name)
-                        (equal? x (string-append name "-scheme")))))
-             (cadr line)
-             (car line)))))
-
 (define (impl-available? cfg spec confirm?)
   (if (find-in-path (cadr spec))
       (or (null? (cddr spec))
+          (not (third spec))
           (conf-get cfg 'skip-version-checks?)
           (let ((version (impl->version (car spec) (third spec))))
             (or (and version (version>=? version (fourth spec)))
@@ -509,8 +488,11 @@
              (files '())
              (lib-dirs '())
              (test test)
-             (extracted-tests? #f))
+             (extracted-tests? #f)
+             (seen '()))
       (cond
+       ((and (pair? ls) (member (caar ls) seen))
+        (lp (cdr ls) progs res files lib-dirs test extracted-tests? seen))
        ((pair? ls)
         (let* ((lib+files (extract-library cfg (caar ls)))
                (lib (car lib+files))
@@ -533,7 +515,8 @@
               (delete-duplicates
                (cons (library-path-base (caar ls) name) lib-dirs))
               test
-              extracted-tests?)))
+              extracted-tests?
+              (cons (caar ls) seen))))
        ((pair? progs)
         (lp ls
             (cdr progs)
@@ -544,7 +527,8 @@
             (cons (car progs) files)
             lib-dirs
             test
-            extracted-tests?))
+            extracted-tests?
+            seen))
        ((null? res)
         (die 2 "No packages generated"))
        ((and (not test)
@@ -566,8 +550,9 @@
                      `(inline
                        "run-tests.scm"
                        ,(test-program-from-libraries tests-from-libraries))
-                     #t)
-                 (lp ls progs res files lib-dirs test #t))))
+                     #t
+                     seen)
+                 (lp ls progs res files lib-dirs test #t seen))))
        (else
         (let* ((docs (package-docs cfg spec libs lib-dirs))
                (desc (package-description cfg spec libs docs))
@@ -962,9 +947,12 @@
    ((package-file? (car o))
     (if (not (every package-file? (cdr o)))
         (non-homogeneous))
-    (for-each
-     (lambda (package) (upload-package cfg spec package))
-     o))
+    ;; TODO: include a summary (version, file size, etc.)
+    (if (yes-or-no? cfg "Upload " o " to "
+                    (remote-uri cfg '(command package uri) "/?"))
+      (for-each
+       (lambda (package) (upload-package cfg spec package))
+       o)))
    (else
     (if (any package-file? (cdr o))
         (non-homogeneous))
@@ -973,7 +961,10 @@
            (package (create-package (car spec+files)
                                     (cdr spec+files)
                                     package-file)))
-      (upload-package cfg spec package package-file)))))
+      ;; TODO: include a summary (version, file size, etc.)
+      (if (yes-or-no? cfg "Upload " o " to "
+                      (remote-uri cfg '(command package uri) "/?"))
+        (upload-package cfg spec package package-file))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Remove - removes the listed libraries.
@@ -981,21 +972,39 @@
 ;; Provides a summary of the libraries to remove along with any
 ;; dependencies they have which were not explicitly installed.
 
-(define (warn-delete-file file)
+(define (remove-with-sudo? cfg path)
+  (case (or (conf-get cfg '(command remove use-sudo?))
+            (conf-get cfg '(command upgrade use-sudo?)))
+    ((always) #t)
+    ((never) #f)
+    (else
+     (not (file-is-writable? (path-directory path))))))
+
+(define (remove-file cfg file)
+  (if (remove-with-sudo? cfg file)
+      (system "sudo" "rm" file)
+      (delete-file file)))
+
+(define (remove-directory cfg dir)
+  (if (remove-with-sudo? cfg dir)
+      (system "sudo" "rmdir" dir)
+      (delete-directory dir)))
+
+(define (warn-delete-file cfg file)
   (guard (exn (else (warn "couldn't delete file: " file)))
-    (delete-file file)))
+    (remove-file cfg file)))
 
 (define (delete-library-files impl cfg pkg lib-name)
-  (for-each warn-delete-file (package-installed-files pkg))
-  (warn-delete-file (make-path (get-install-source-dir impl cfg)
-                               (get-package-meta-file cfg pkg)))
+  (for-each (lambda (f) (warn-delete-file cfg f)) (package-installed-files pkg))
+  (warn-delete-file cfg (make-path (get-install-source-dir impl cfg)
+                                   (get-package-meta-file cfg pkg)))
   (cond
    ((package->path cfg pkg)
     => (lambda (path)
          (let ((dir (make-path (get-install-source-dir impl cfg) path)))
            (if (and (file-directory? dir)
                     (= 2 (length (directory-files dir))))
-               (delete-directory dir)))))))
+               (remove-directory cfg dir)))))))
 
 (define (command/remove cfg spec . args)
   (let* ((impls (conf-selected-implementations cfg))
@@ -1023,13 +1032,18 @@
             (map car lib-names+pkgs)
             (map cdr lib-names+pkgs)))
 
-;; faster than (length (regexp-extract re str))
-(define (regexp-count re str)
-  (regexp-fold re (lambda (from md str acc) (+ acc 1)) 0 str))
+(define (string-count-word str word)
+  (let lp ((sc (string-cursor-start str)) (count 0))
+    (let ((sc2 (string-contains str word sc)))
+      (if sc2
+          (lp (string-cursor-next str sc2) (+ count 1))
+          count))))
 
 (define (count-in-sexp x keywords)
-  (regexp-count `(word (w/nocase (or ,@keywords)))
-                (write-to-string x)))
+  (let ((s (string-downcase (write-to-string x))))
+    (fold (lambda (k sum) (+ sum (string-count-word s k)))
+          0
+          (map string-downcase keywords))))
 
 (define (extract-matching-libraries cfg repo keywords)
   (define (library-score lib)
@@ -1139,15 +1153,15 @@
          (local-tmp (string-append local-path ".tmp."
                                    (number->string (current-second)) "-"
                                    (number->string (current-process-id))))
-         (repo-str (utf8->string (resource->bytevector repo-uri)))
+         (repo-str (utf8->string (resource->bytevector cfg repo-uri)))
          (repo (guard (exn (else #f))
                  (let ((repo (read (open-input-string repo-str))))
                    `(,(car repo) (url ,repo-uri) ,@(cdr repo))))))
     (cond
      ((not (valid-repository? repo))
-      (die 2 "not a valid repository: " repo-uri))
+      (warn "not a valid repository: " repo-uri repo))
      ((not (create-directory* local-dir))
-      (die 2 "can't create directory: " local-dir))
+      (warn "can't create directory: " local-dir))
      (else
       (guard (exn (else (die 2 "couldn't write repository")))
         (call-with-output-file local-tmp
@@ -1287,6 +1301,16 @@
       (string->number (process->string '(csi -p "(##sys#fudge 42)")))
       8))
 
+(define (get-chicken-repo-path)
+  (let ((release (string-trim (process->string '(csi -release))
+                              char-whitespace?)))
+    (string-trim
+     (if (string-prefix? "4." release)
+         (process->string '(csi -p "(repository-path)"))
+         (process->string
+          '(csi -R chicken.platform -p "(car (repository-path))")))
+     char-whitespace?)))
+
 (define (get-install-dirs impl cfg)
   (define (guile-eval expr)
     (guard (exn (else #f))
@@ -1304,9 +1328,7 @@
            (cons share-dir (delete share-dir dirs))
            dirs)))
     ((chicken)
-     (let ((dir (string-trim
-                 (process->string '(csi -p "(repository-path)"))
-                 char-whitespace?)))
+     (let ((dir (get-chicken-repo-path)))
        (list
         (if (file-exists? dir)  ; repository-path should always exist
             dir
@@ -1352,6 +1374,24 @@
                       "share/snow"
                       impl)))))
 
+(define (get-install-library-dirs impl cfg)
+  (case impl
+    ((chibi)
+     (let* ((dirs
+             (reverse
+              (cond-expand
+               (chibi (eval '(current-module-path) (environment '(chibi))))
+               (else (process->sexp
+                      '(chibi-scheme -q -p "(current-module-path)"))))))
+            (lib-dir (find (lambda (d) (string-contains d "/lib")) dirs)))
+       (if lib-dir
+           (cons lib-dir (delete lib-dir dirs))
+           dirs)))
+    (else
+     (list (make-path (or (conf-get cfg 'install-prefix) "/usr/local")
+                      "lib"
+                      impl)))))
+
 (define (scheme-script-command impl cfg)
   (or (and (eq? impl 'chibi) (conf-get cfg 'chibi-path))
       (let* ((prog (cond ((conf-get cfg 'scheme-script))
@@ -1525,7 +1565,9 @@
           ((and (equal? "meta" (path-extension file))
                 (guard (exn (else #f))
                   (let ((pkg (call-with-input-file file read)))
-                    (and (package? pkg) pkg))))
+                    (and (package? pkg)
+                         (every file-exists? (package-installed-files pkg))
+                         pkg))))
            => (lambda (pkg)
                 (append
                  (map
@@ -1630,7 +1672,7 @@
     (car (get-install-dirs impl cfg)))
    ((conf-get cfg 'install-prefix)
     => (lambda (prefix) (make-path prefix "lib" impl)))
-   (else (make-path "/usr/local/lib" impl))))
+   (else (car (get-install-library-dirs impl cfg)))))
 
 (define (get-install-binary-dir impl cfg)
   (cond
@@ -2073,7 +2115,7 @@
     (install-file cfg (make-path dir src) dest)))
 
 (define (fetch-package cfg url)
-  (resource->bytevector url))
+  (resource->bytevector cfg url))
 
 (define (path-strip-top file)
   (let ((pos (string-find file #\/)))
