@@ -1,18 +1,16 @@
-;; Copyright (c) 2010-2014 Alex Shinn. All rights reserved. 
+;; Copyright (c) 2010-2020 Alex Shinn. All rights reserved.
 ;; BSD-style license: http://synthcode.com/license.txt
 
-;;> Simple testing framework adapted from the Chicken \scheme{test}
-;;> module.
+;;> Simple but extensible testing framework with advanced reporting.
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; list utilities
 
-;; Simplified version of SRFI-1 every.
-(define (every pred ls)
-  (or (null? ls)
-      (if (null? (cdr ls))
-          (pred (car ls))
-          (if (pred (car ls)) (every pred (cdr ls)) #f))))
+;; Simplified version of SRFI-1 any.
+(define (any pred ls)
+  (and (pair? ls)
+       (or (pred (car ls))
+           (any pred (cdr ls)))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; exception utilities
@@ -44,11 +42,28 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; test interface
 
+;;> \section{Testing}
+
 ;;> \macro{(test [name] expect expr)}
 
-;;> Evaluate \var{expr} and check that it is \scheme{equal?}
-;;> to \var{expect}.  \var{name} is used in reporting, and
-;;> defaults to a printed summary of \var{expr}.
+;;> The primary interface to testing.  Evaluate \var{expr} and check
+;;> that it is equal to \var{expect}, and report the result, using
+;;> \var{name} or a printed summary of \var{expr}.
+;;>
+;;> If used inside a group this will contribute to the overall group
+;;> reporting, but can be used standalone:
+;;>
+;;> \example{(test 4 (+ 2 2))}
+;;> \example{(test "add two and two" 4 (+ 2 2))}
+;;> \example{(test 3 (+ 2 2))}
+;;> \example{(test 4 (+ 2 "2"))}
+;;>
+;;> The equality comparison is made with
+;;> \scheme{current-test-comparator}, defaulting to
+;;> \scheme{test-equal?}, which is the same as \scheme{equal?} but
+;;> more permissive on floating point comparisons).  Returns the
+;;> status of the test (one of the symbols \scheme{'PASS},
+;;> \scheme{'FAIL}, \scheme{'SKIP}, \scheme{'ERROR}).
 
 (define-syntax test
   (syntax-rules (quote)
@@ -130,10 +145,12 @@
      (test-syntax-error 'test-error "1 or 2 arguments required"
                         (test a ...)))))
 
-;; TODO: Extract interesting variables so we can show their values on
-;; failure.
+;;> Low-level macro to pass alist info to the underlying \var{test-run}.
+
 (define-syntax test-propagate-info
   (syntax-rules ()
+    ;; TODO: Extract interesting variables so we can show their values
+    ;; on failure.  Vars are empty for now.
     ((test-propagate-info name expect expr info)
      (test-vars () name expect expr info))))
 
@@ -148,27 +165,68 @@
                  (var-values . ,(list vars ...))
                  (key . val) ...)))))
 
-;;> \macro{(test-exit)}
+;;> The procedural interface to testing.  \var{expect} and \var{expr}
+;;> should be thunks, and \var{info} is an alist of properties used in
+;;> test reporting.
 
-;;> Exits with a failure status if any tests have failed,
-;;> and a successful status otherwise.
+(define (test-run expect expr info)
+  (let ((info (test-expand-info info)))
+    (if (and (cond ((current-test-group)
+                    => (lambda (g) (not (test-group-ref g 'skip-group?))))
+                   (else #t))
+             (or (and (not (any (lambda (f) (f info)) (current-test-removers)))
+                      (or (pair? (current-test-removers))
+                          (null? (current-test-filters))))
+                 (any (lambda (f) (f info)) (current-test-filters))))
+        ((current-test-applier) expect expr info)
+        ((current-test-skipper) info))))
 
-(define (test-exit)
-  (exit (zero? (test-failure-count))))
+;;> Returns true if either \scheme{(equal? expect res)}, or
+;;> \var{expect} is inexact and \var{res} is within
+;;> \scheme{current-test-epsilon} of \var{expect}.
+
+(define (test-equal? expect res)
+  (or (equal? expect res)
+      (if (real? expect)
+          (and (inexact? expect)
+               (real? res)
+               ;; tests which expect an inexact value can
+               ;; accept an equivalent exact value
+               ;; (inexact? res)
+               (approx-equal? expect res (current-test-epsilon)))
+          (and (complex? res)
+               (complex? expect)
+               (test-equal? (real-part expect) (real-part res))
+               (test-equal? (imag-part expect) (imag-part res))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; group interface
 
-;;> Wraps \var{body} as a single test group, which can be filtered
-;;> and summarized separately.
+;;> \section{Test Groups}
+
+;;> Tests can be collected in groups for separate reporting, filtering
+;;> and for catching exceptions outside of a test case.
+
+;;> Wraps \var{body} as a single test group, which can be filtered and
+;;> summarized separately.  The \var{body} is arbitrary Scheme code,
+;;> and tests run within its dynamic extent will be associated with
+;;> the group.  If an uncaught exception is raised outside of a test
+;;> case, it will cause the whole group to fail with an error status.
+
+;;> \example{
+;;> (test-group "pi"
+;;>   (test 3.14159 (acos -1))
+;;>   (test 3 (acos -1))
+;;>   (test 3.14159 (acos "-1")))
+;;> }
 
 (define-syntax test-group
   (syntax-rules ()
     ((_ name-expr body ...)
      (let ((name name-expr)
            (old-group (current-test-group)))
-       (if (not (string? name))
-           (error "a name is required, got " 'name-expr name))
+       (when (not (string? name))
+         (error "a name is required, got " 'name-expr name))
        (test-begin name)
        (guard
            (exn
@@ -182,6 +240,78 @@
        (test-end name)
        (current-test-group old-group)))))
 
+;;> Begin testing a new group until the closing \scheme{(test-end)}.
+
+(define (test-begin . o)
+  (let* ((name (if (pair? o) (car o) ""))
+         (parent (current-test-group))
+         (group (make-test-group name parent)))
+    ;; include a newline if we are directly nested in a parent with no
+    ;; tests yet
+    (when (and parent
+               (zero? (test-group-ref parent 'subgroups-count 0))
+               (not (test-group-ref parent 'verbose)))
+      (newline))
+    ;; header
+    (cond
+     ((test-group-ref group 'skip-group?)
+      (display (make-string (or (test-group-indent-width group) 0) #\space))
+      (display (strikethrough (bold (string-append name ":"))))
+      (display " SKIP"))
+     ((test-group-ref group 'verbose)
+      (display
+       (test-header-line
+        (string-append "testing " name)
+        (or (test-group-indent-width group) 0))))
+     (else
+      (display
+       (string-append
+        (make-string (or (test-group-indent-width group) 0)
+                     #\space)
+        (bold (string-append name ": "))))))
+    ;; set the current test group
+    (current-test-group group)))
+
+;;> Ends testing group introduced with \scheme{(test-begin)}, and
+;;> summarizes the results.  The \var{name} is optional, but if
+;;> present should match the corresponding \scheme{test-begin} name,
+;;> or a warning is printed.
+
+(define (test-end . o)
+  (let ((name (and (pair? o) (car o))))
+    (cond
+     ((current-test-group)
+      => (lambda (group)
+           (when (and name (not (equal? name (test-group-name group))))
+             (warning "mismatched test-end:" name (test-group-name group)))
+           (let ((parent (test-group-ref group 'parent)))
+             (when (and (test-group-ref group 'skip-group?)
+                        (zero? (test-group-ref group 'subgroups-count 0)))
+               (newline))
+             ;; only report if there's something to say
+             ((current-test-group-reporter) group)
+             (when parent
+               (test-group-inc! parent 'subgroups-count)
+               (cond
+                ((test-group-ref group 'skip-group?)
+                 (test-group-inc! parent 'subgroups-skip))
+                ((and (zero? (test-group-ref group 'FAIL 0))
+                      (zero? (test-group-ref group 'ERROR 0))
+                      (= (test-group-ref group 'subgroups-pass 0)
+                         (test-group-ref group 'subgroups-count 0)))
+                 (test-group-inc! parent 'subgroups-pass))))
+             (current-test-group parent)
+             group))))))
+
+;;> Exits with a failure status if any tests have failed,
+;;> and a successful status otherwise.
+
+(define (test-exit)
+  (when (current-test-group)
+    (warning "calling test-exit with unfinished test group:"
+             (test-group-name (current-test-group))))
+  (exit (zero? (test-failure-count))))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; utilities
 
@@ -192,6 +322,8 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; test-group representation
 
+;;> \section{Accessors}
+
 ;; (name (prop value) ...)
 (define (make-test-group name . o)
   (let ((parent (and (pair? o) (car o)))
@@ -208,23 +340,39 @@
     (test-group-set!
      group
      'skip-group?
-     (or (and parent (test-group-ref parent 'skip-group?))
-         (not (every (lambda (f) (f group)) (current-test-group-filters)))))
+     (and (or (and parent (test-group-ref parent 'skip-group?))
+              (any (lambda (f) (f group)) (current-test-group-removers))
+              (and (null? (current-test-group-removers))
+                   (pair? (current-test-group-filters))))
+          (not (any (lambda (f) (f group)) (current-test-group-filters)))))
     group))
 
-(define test-group-name car)
+;;> Returns the name of a test group info object.
+
+(define (test-group-name group) (car group))
+
+;;> Returns the value of a \var{field} in a test var{group} info
+;;> object.  \var{field} should be a symbol, and predefined fields
+;;> include \scheme{parent}, \scheme{verbose}, \scheme{level},
+;;> \scheme{start-time}, \scheme{skip-group?}, \scheme{count},
+;;> \scheme{total-pass}, \scheme{total-fail}, \scheme{total-error}.
 
 (define (test-group-ref group field . o)
   (if group
       (apply assq-ref (cdr group) field o)
       (and (pair? o) (car o))))
 
+;;> Sets the value of a \var{field} in a test \var{group} info object.
+
 (define (test-group-set! group field value)
   (cond
    ((assq field (cdr group))
     => (lambda (x) (set-cdr! x value)))
    (else (set-cdr! group (cons (cons field value) (cdr group))))))
 
+;;> Increments the value of a \var{field} in a test \var{group} info
+;;> object by \var{amount}, defaulting to 1.
+
 (define (test-group-inc! group field . o)
   (let ((amount (if (pair? o) (car o) 1)))
     (cond
@@ -232,6 +380,9 @@
       => (lambda (x) (set-cdr! x (+ amount (cdr x)))))
      (else (set-cdr! group (cons (cons field amount) (cdr group)))))))
 
+;;> Updates a \var{field} in a test group info object by consing
+;;> \var{value} onto it.
+
 (define (test-group-push! group field value)
   (cond
    ((assq field (cdr group))
@@ -369,20 +520,11 @@
           ,@info)
         info)))
 
-(define (test-run expect expr info)
-  (let ((info (test-expand-info info)))
-    (if (and (cond ((current-test-group)
-                    => (lambda (g) (not (test-group-ref g 'skip-group?))))
-                   (else #t))
-             (every (lambda (f) (f info)) (current-test-filters)))
-        ((current-test-applier) expect expr info)
-        ((current-test-skipper) info))))
-
 (define (test-default-applier expect expr info)
   (let* ((group (current-test-group))
          (indent (and group (test-group-indent-width group))))
     (cond
-     ((test-group-ref group 'verbose)
+     ((or (not group) (test-group-ref group 'verbose))
       (if (and indent (positive? indent))
           (display (make-string indent #\space)))
       (test-print-name info indent)))
@@ -397,7 +539,7 @@
       (guard
           (exn
            (else
-            ((current-test-handler)
+            ((current-test-reporter)
              (if (assq-ref info 'expect-error) 'PASS 'ERROR)
              (append `((exception . ,exn)) info))))
         (let ((res (expr)))
@@ -409,10 +551,10 @@
                      'PASS
                      'FAIL))
                 (info `((result . ,res) (expected . ,expect-val) ,@info)))
-            ((current-test-handler) status info)))))))
+            ((current-test-reporter) status info)))))))
 
 (define (test-default-skipper info)
-  ((current-test-handler) 'SKIP info))
+  ((current-test-reporter) 'SKIP info))
 
 (define (test-status-color status)
   (case status
@@ -426,12 +568,23 @@
 
 (define (test-status-code status)
   ((test-status-color status)
+   ;; alternatively: ❗, ✗, ‒, ✓
+   ;; unfortunately, these have ambiguous width
    (case status
      ((ERROR) "!")
      ((FAIL) "x")
      ((SKIP) "-")
      (else "."))))
 
+(define (display-expected/actual expected actual)
+  (let* ((e-str (write-to-string expected))
+         (a-str (write-to-string actual))
+         (diff (diff e-str a-str read-char)))
+    (write-string "expected ")
+    (write-string (edits->string/color (car diff) (car (cddr diff)) 1))
+    (write-string " but got ")
+    (write-string (edits->string/color (cadr diff) (car (cddr diff)) 2))))
+
 (define (test-print-explanation indent status info)
   (cond
    ((eq? status 'ERROR)
@@ -448,8 +601,8 @@
     (write (assq-ref info 'result)) (newline))
    ((eq? status 'FAIL)
     (display indent)
-    (display "expected ") (write (assq-ref info 'expected))
-    (display " but got ") (write (assq-ref info 'result)) (newline)))
+    (display-expected/actual (assq-ref info 'expected) (assq-ref info 'result))
+    (newline)))
   ;; print variables
   (cond
    ((and (memq status '(FAIL ERROR)) (assq-ref info 'var-names))
@@ -501,14 +654,13 @@
   ;; display line, source and values info
   (test-print-source indent status info))
 
-(define (test-print-header-line str . indent)
+(define (test-header-line str . indent)
   (let* ((header (string-append
                   (make-string (if (pair? indent) (car indent) 0) #\space)
                   "-- " str " "))
          (len (string-length header)))
-    (display (bold header))
-    (display (make-string (max 0 (- (current-column-width) len)) #\-))
-    (newline)))
+    (string-append (bold header)
+                   (make-string (max 0 (- (current-column-width) len)) #\-))))
 
 (define (test-default-handler status info)
   (define indent
@@ -540,8 +692,8 @@
    ((or (eq? status 'FAIL) (eq? status 'ERROR))
     (test-failure-count (+ 1 (test-failure-count)))))
   (cond
-   ((eq? status 'SKIP))
-   ((test-group-ref (current-test-group) 'verbose)
+   ((or (not (current-test-group))
+        (test-group-ref (current-test-group) 'verbose))
     ;; display status
     (display "[")
     (if (not (eq? status 'ERROR)) (display " ")) ; pad
@@ -549,6 +701,7 @@
     (display "]")
     (newline)
     (test-print-failure indent status info))
+   ((eq? status 'SKIP))
    (else
     (display (test-status-code status))
     (cond
@@ -579,6 +732,8 @@
          (err (+ base-err (or (test-group-ref group 'total-error) 0)))
          (count (+ pass fail err))
          (subgroups-count (or (test-group-ref group 'subgroups-count) 0))
+         (subgroups-skip (or (test-group-ref group 'subgroups-skip) 0))
+         (subgroups-run (- subgroups-count subgroups-skip))
          (subgroups-pass (or (test-group-ref group 'subgroups-pass) 0))
          (indent (make-string (or (test-group-indent-width group) 0) #\space)))
     (if (and (not (test-group-ref group 'verbose))
@@ -632,22 +787,23 @@
            (apply test-print-failure failure))
          (reverse (or (test-group-ref group 'failures) '())))))
       (cond
-       ((positive? subgroups-count)
+       ((positive? subgroups-run)
         (display indent)
         (display
-         ((if (= subgroups-pass subgroups-count)
+         ((if (= subgroups-pass subgroups-run)
               green (lambda (x) x))
           (string-append
            (number->string subgroups-pass) " out of "
-           (number->string subgroups-count)
-           (percent subgroups-pass subgroups-count))))
+           (number->string subgroups-run)
+           (percent subgroups-pass subgroups-run))))
         (display (plural " subgroup" subgroups-pass))
         (display " passed.\n")))))
     (cond
      ((test-group-ref group 'verbose)
-      (test-print-header-line
-       (string-append "done testing " (or (test-group-name group) ""))
-       (or (test-group-indent-width group) 0))
+      (display
+       (test-header-line
+        (string-append "done testing " (or (test-group-name group) ""))
+        (or (test-group-indent-width group) 0)))
       (newline)))
     (cond
      ((test-group-ref group 'parent)
@@ -657,85 +813,62 @@
            (test-group-inc! parent 'total-fail fail)
            (test-group-inc! parent 'total-error err))))))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-
-(define (test-equal? expect res)
-  (or (equal? expect res)
-      (if (real? expect)
-          (and (inexact? expect)
-               (real? res)
-               (inexact? res)
-               (approx-equal? expect res (current-test-epsilon)))
-          (and (complex? res)
-               (complex? expect)
-               (test-equal? (real-part expect) (real-part res))
-               (test-equal? (imag-part expect) (imag-part res))))))
-
-;;> Begin testing a new group until the closing \scheme{(test-end)}.
-
-(define (test-begin . o)
-  (let* ((name (if (pair? o) (car o) ""))
-         (parent (current-test-group))
-         (group (make-test-group name parent)))
-    (cond
-     ((and parent
-           ;; (zero? (test-group-ref parent 'count 0))
-           (zero? (test-group-ref parent 'subgroups-count 0)))
-      (newline)))
-    (cond
-     ((test-group-ref group 'verbose)
-      (test-print-header-line
-       (string-append "testing " name)
-       (or (test-group-indent-width group) 0)))
-     (else
-      (display
-       (make-string (or (test-group-indent-width group) 0)
-                    #\space))
-      (display (bold (string-append name ": ")))))
-    (current-test-group group)))
-
-;;> Ends testing group introduced with \scheme{(test-begin)}, and
-;;> summarizes the results.
-
-(define (test-end . o)
-  (cond
-   ((current-test-group)
-    => (lambda (group)
-         (if (and (pair? o) (not (equal? (car o) (test-group-name group))))
-             (warning "mismatched test-end:" (car o) (test-group-name group)))
-         (let ((parent (test-group-ref group 'parent)))
-           (cond
-            ((not (test-group-ref group 'skip-group?))
-             ;; only report if there's something to say
-             ((current-test-group-reporter) group)
-             (cond
-              (parent
-               (test-group-inc! parent 'subgroups-count)
-               (cond
-                ((and (zero? (test-group-ref group 'FAIL 0))
-                      (zero? (test-group-ref group 'ERROR 0))
-                      (= (test-group-ref group 'subgroups-pass 0)
-                         (test-group-ref group 'subgroups-count 0)))
-                 (test-group-inc! parent 'subgroups-pass)))))))
-           (current-test-group parent)
-           group)))))
-
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; parameters
 
+;;> \section{Parameters}
+
+;;> The current test group as started by \scheme{test-group} or
+;;> \scheme{test-begin}.
+
 (define current-test-group (make-parameter #f))
+
+;;> If true, show more verbose output per test.  Inferred from the
+;;> environment variable TEST_VERBOSE.
+
 (define current-test-verbosity
   (make-parameter
    (cond ((get-environment-variable "TEST_VERBOSE")
           => (lambda (s) (not (member s '("" "0")))))
          (else #f))))
+
+;;> The epsilon used for floating point comparisons.
+
 (define current-test-epsilon (make-parameter 1e-5))
+
+;;> The underlying comparator used in testing, defaults to
+;;> \scheme{test-equal?}.
+
 (define current-test-comparator (make-parameter test-equal?))
+
+;;> The test applier - what we do with non-skipped tests.  Takes the
+;;> same signature as \scheme{test-run}, should be responsible for
+;;> evaluating the thunks, determining the status of the test, and
+;;> passing this information to \scheme{current-test-reporter}.
+
 (define current-test-applier (make-parameter test-default-applier))
-(define current-test-handler (make-parameter test-default-handler))
+
+;;> The test skipper - what we do with non-skipped tests.  This should
+;;> not evaluate the thunks and simply pass off to
+;;> \scheme{current-test-reporter}.
+
 (define current-test-skipper (make-parameter test-default-skipper))
+
+;;> Takes two arguments, the symbol status of the test and the info
+;;> alist.  Reports the result of the test and updates bookkeeping in
+;;> the current test group for reporting.
+
+(define current-test-reporter (make-parameter test-default-handler))
+
+;;> Takes one argument, a test group, and prints a summary of the test
+;;> results for that group.
+
 (define current-test-group-reporter
   (make-parameter test-default-group-reporter))
+
+;;> A running count of all test failures and errors across all groups
+;;> (and threads).  Used by \scheme{test-exit}.
+
 (define test-failure-count (make-parameter 0))
 
 (define test-first-indentation
@@ -759,35 +892,94 @@
           (else #f))))
 
 (define (string->group-matcher str)
-  (lambda (group) (string-search str (car group))))
+  (lambda (group) (string-search str (test-group-name group))))
 
-(define (getenv-filter-list proc name . o)
+;; simplified version from SRFI 130
+(define (string-split str ch)
+  (let ((end (string-length str)))
+    (let lp ((from 0) (to 0) (res '()))
+      (cond
+       ((>= to end)
+        (reverse (if (> to from) (cons (substring str from to) res) res)))
+       ((eqv? ch (string-ref str to))
+        (lp (+ to 1) (+ to 1) (cons (substring str from to) res)))
+       (else
+        (lp from (+ to 1) res))))))
+
+(define (getenv-filter-list proc name)
   (cond
    ((get-environment-variable name)
     => (lambda (s)
-         (guard
-             (exn
-              (else
-               (warning
-                (string-append "invalid filter '" s
-                               "' from environment variable: " name))
-               (print-exception exn (current-error-port))
-               '()))
-           (let ((f (proc s)))
-             (list (if (and (pair? o) (car o))
-                       (lambda (x) (not (f x)))
-                       f))))))
+         (let lp ((ls (string-split s #\,))
+                  (res '()))
+           (cond
+            ((null? ls) (reverse res))
+            (else
+             (let* ((s (car ls))
+                    (f (guard
+                           (exn
+                            (else
+                             (warning
+                              (string-append "invalid filter '" s
+                                             "' from environment variable: "
+                                             name))
+                             (print-exception exn (current-error-port))
+                             #f))
+                         (proc s))))
+               (lp (cdr ls) (if f (cons f res) res))))))))
    (else '())))
 
-(define current-test-filters
-  (make-parameter
-   (append (getenv-filter-list string->info-matcher "TEST_FILTER")
-           (getenv-filter-list string->info-matcher "TEST_REMOVE" #t))))
-
 (define current-test-group-filters
   (make-parameter
-   (append (getenv-filter-list string->group-matcher "TEST_GROUP_FILTER")
-           (getenv-filter-list string->group-matcher "TEST_GROUP_REMOVE" #t))))
+   (getenv-filter-list string->group-matcher "TEST_GROUP_FILTER")))
+
+(define current-test-group-removers
+  (make-parameter
+   (getenv-filter-list string->group-matcher "TEST_GROUP_REMOVE")))
+
+;;> Parameters controlling which test groups are skipped.  Each
+;;> parameter is a list of procedures of one argument, a test group
+;;> info, which can be queried with \var{test-group-name} and
+;;> \var{test-group-ref}.  Analogous to SRFI 1, a filter selects a
+;;> group for inclusion and a removers for exclusion.  The defaults
+;;> are set automatically from the environment variables
+;;> TEST_GROUP_FILTER and TEST_GROUP_REMOVE, which should be
+;;> comma-delimited lists of strings which are checked for a substring
+;;> match in the test group name.  A test group is skipped if it does
+;;> not match any filter and:
+;;> \itemlist[
+;;> \item{its parent group is skipped, or}
+;;> \item{it matches a remover, or}
+;;> \item{no removers are specified but some filters are}
+;;> ]
+;;/
+
+(define current-test-filters
+  (make-parameter (getenv-filter-list string->info-matcher "TEST_FILTER")))
+
+(define current-test-removers
+  (make-parameter (getenv-filter-list string->info-matcher "TEST_REMOVE")))
+
+;;> Parameters controlling which tests are skipped.  Each parameter is
+;;> a list of procedures of one argument, a test info alist, which can
+;;> be queried with \scheme{test-get-name!} or \scheme{assq}.
+;;> Analogous to SRFI 1, a filter selects a test for inclusion and a
+;;> removers for exclusion.  The defaults are set automatically from
+;;> the environment variables TEST_FILTER and TEST_REMOVE, which
+;;> should be comma-delimited lists of strings which are checked for a
+;;> substring match in the test name.  A test is skipped if its group
+;;> is skipped, or if it does not match a filter and:
+;;> \itemlist[
+;;> \item{it matches a remover, or}
+;;> \item{no removers are specified but some filters are}
+;;> ]
+;;/
+
+;;> Parameter controlling the current column width for test output,
+;;> can be set from the environment variable TEST_COLUMN_WIDTH,
+;;> otherwise defaults to 78.  For portability of implementation (and
+;;> resulting output), does not attempt to use termios to determine
+;;> the actual available width.
 
 (define current-column-width
   (make-parameter
