@@ -1,9 +1,9 @@
 
-(define fl= =)
-(define fl< <)
-(define fl> >)
-(define fl<= <=)
-(define fl>= >=)
+(define fl=? =)
+(define fl<? <)
+(define fl>? >)
+(define fl<=? <=)
+(define fl>=? >=)
 (define flodd? odd?)
 (define fleven? even?)
 (define (flunordered? x y) (or (flnan? x) (flnan? y)))
